from zealous.exceptions import BaseError


class RPCError(BaseError):
    message = "An unexpected RPCError occurred ({details})"


class RPCServerNotStarted(RPCError):
    message = "RPC server was not started"


class RPCServerAlreadyStarted(RPCError):
    message = "RPC server was already started"


class RPCServerInvalidConfig(RPCError):
    message = "RPC server configuration is invalid ({reason})"


class RPCServerInterfaceNotImplemented(RPCError):
    message = (
        "RPC server interface {interface} for resource {resource} not implemented."
    )

    def __init__(self, *, resource: str, interface: str) -> None:
        super().__init__(resource=resource, interface=interface)
        self.resource = resource
        self.interface = interface


class RPCRemoteError(RPCError):
    """An error raised in RPC backend and re-raised by the client."""

    message = "A remote exception occurred ({exc_type}: {exc_message})"

    def __init__(self, *, exc_type: str, exc_message: str) -> None:
        super().__init__(exc_type=exc_type, exc_message=exc_message)
        self.exc_type = exc_type
        self.exc_message = exc_message



from __future__ import annotations

import sys

from pydantic import BaseModel as Schema

from zealous import exceptions


__all__ = (
    "ErrorSchema",
    "Schema",
)


class NoExceptionCurrentlyBeingHandled(exceptions.BaseError):
    message = "No exception is currently being handled."


class ErrorSchema(Schema):
    type: str
    message: str

    @classmethod
    def from_exc_info(cls) -> ErrorSchema:
        """Generate error for the exception that is currently being handled."""
        type_, value, _ = sys.exc_info()
        if type_ and value:
            return cls(type=type_.__name__, message=str(value))
        raise NoExceptionCurrentlyBeingHandled



from __future__ import annotations

import abc
import asyncio
import logging

from dataclasses import dataclass
from typing import Any
from typing import Generic
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast
from typing import overload

from zealous import ErrorSchema
from zealous import Schema
from zealous.base import App
from zealous.base import Boundary
from zealous.base import BoundaryT
from zealous.base import Interface
from zealous.base import InterfaceNoBody
from zealous.base import Resource
from zealous.typing import SchemaInContraT
from zealous.typing import SchemaOutCoT

from . import exceptions


__all__ = (
    "RPCClientAppMixin",
    "RPCClientConf",
    "RPCServerAppMixin",
    "RPCServerConf",
)

LOG = logging.getLogger(__name__)


DriverContextT = TypeVar("DriverContextT")


class RPCContext(Generic[DriverContextT]):
    def __init__(
        self, *, resource: str, interface: str, driver_context: DriverContextT,
    ) -> None:
        self.resource = resource
        self.interface = interface
        self.driver_context = driver_context


class BaseRPCDriver(Generic[DriverContextT], metaclass=abc.ABCMeta):
    def make_context(
        self, resource: str, interface: str,
    ) -> RPCContext[DriverContextT]:
        return RPCContext(
            resource=resource,
            interface=interface,
            driver_context=self.make_driver_context(),
        )

    @abc.abstractmethod
    def make_driver_context(self) -> DriverContextT:
        """Hook for attaching driver_context to RPCMessageContext."""

    @abc.abstractmethod
    async def push_rpc_message_in(
        self, context: RPCContext[DriverContextT], body: Optional[Schema]
    ) -> None:
        ...

    @abc.abstractmethod
    async def pop_rpc_message_in(
        self,
    ) -> Tuple[RPCContext[DriverContextT], Optional[Schema]]:
        ...

    @abc.abstractmethod
    async def push_rpc_message_out(
        self,
        context: RPCContext[DriverContextT],
        response_body: Union[Schema, Tuple[Schema, ...], None],
    ) -> None:
        ...

    @abc.abstractmethod
    async def pop_rpc_message_out(
        self, context: RPCContext[DriverContextT]
    ) -> Union[Schema, Tuple[Schema, ...], None]:
        ...


@dataclass
class RPCClientConf:
    driver: BaseRPCDriver[Any]


class RPCClient:
    def __init__(self, conf: RPCClientConf, boundary: Boundary) -> None:
        """Hook up RPCResourceClient as per the boundary's specification."""
        for attr, resource in boundary.__resources__.items():
            setattr(self, attr, RPCResourceClient(conf, resource, attr))


class RPCResourceClient:
    """RPC client for a specific resource."""

    def __init__(
        self, conf: RPCClientConf, resource: Resource, resource_name: str
    ) -> None:
        """Hook up RPCInterfaceClient as per resource definition."""
        for attr, interface in resource.__interfaces__.items():
            setattr(
                self, attr, RPCInterfaceClient(conf, interface, resource_name, attr),
            )


class RPCInterfaceClient(Interface[SchemaInContraT, SchemaOutCoT]):
    """RPC client for a specific resource interface."""

    def __init__(
        self,
        conf: RPCClientConf,
        interface: Union[Interface[Any, Any], InterfaceNoBody[Any]],
        resource_name: str,
        interface_name: str,
    ):
        self._conf = conf
        self._interface = interface
        self._resource_name = resource_name
        self._interface_name = interface_name

    async def __call__(self, body: SchemaInContraT = None) -> SchemaOutCoT:
        context = self._conf.driver.make_context(
            self._resource_name, self._interface_name
        )
        await self._conf.driver.push_rpc_message_in(context, body)
        response_body = await self._conf.driver.pop_rpc_message_out(context)
        if isinstance(response_body, ErrorSchema):
            raise exceptions.RPCRemoteError(
                exc_type=response_body.type, exc_message=response_body.message
            )
        # if the response_body is not an error, the type is enforced by the interface
        return cast(SchemaOutCoT, response_body)


class RPCClientAppMixin(App[BoundaryT]):
    """Provides RPC Client implementation for Boundary."""

    def __init__(self, boundary: BoundaryT) -> None:
        super().__init__(boundary)
        self._rpc_client: Optional[RPCClient] = None

    @property
    @abc.abstractmethod
    def __rpc_client__(self) -> RPCClientConf:
        ...

    @property
    def rpc_client(self) -> BoundaryT:
        if self._rpc_client is None:
            self._rpc_client = RPCClient(self.__rpc_client__, self._boundary)
        # for all intents and purposes, this is just an implementation of the boundary
        return cast(BoundaryT, self._rpc_client)


@dataclass
class RPCServerConf:
    driver: BaseRPCDriver[Any]
    impl: Boundary
    n_workers: int = 1


class RPCServerAppMixin(App[BoundaryT]):
    """Provides RPC server implementation for Boundary."""

    def __init__(self, boundary: BoundaryT) -> None:
        super().__init__(boundary)
        self._rpc_server: Optional[RPCServer] = None

    @property
    @abc.abstractmethod
    def __rpc_server__(self) -> RPCServerConf:
        ...

    @property
    def rpc_server(self) -> RPCServer:
        if self._rpc_server is None:
            self._rpc_server = RPCServer(self.__rpc_server__)
        return self._rpc_server


HandlerReturns = Union[Schema, Tuple[Schema, ...], None]


class RPCServer:
    def __init__(self, conf: RPCServerConf) -> None:
        self._conf = conf

        if self._conf.n_workers < 1:
            raise exceptions.RPCServerInvalidConfig(
                reason="RPCServer n_workers must be >= 1"
            )
        self._worker_tasks: Optional[List[asyncio.Task[None]]] = None

    async def start(self) -> None:
        if self._worker_tasks:
            raise exceptions.RPCServerAlreadyStarted
        self._worker_tasks = [
            asyncio.create_task(self.loop()) for _ in range(self._conf.n_workers)
        ]

    async def stop(self) -> None:
        if not self._worker_tasks:
            raise exceptions.RPCServerNotStarted
        for task in self._worker_tasks:
            task.cancel()
        await asyncio.gather(*self._worker_tasks)
        self._worker_tasks = None

    async def loop(self) -> None:
        while True:
            try:
                await self.loop_once()
            except asyncio.CancelledError:
                LOG.debug("Server loop cancelled. Leaving.")
                break
            except Exception:  # pylint: disable=broad-except
                LOG.exception("Error caught in server loop. Continuing...")

    async def loop_once(self) -> None:
        context, body = await self._conf.driver.pop_rpc_message_in()
        try:
            response_body = await self.delegate_rpc_message(context, body)
        except Exception:  # pylint: disable=broad-except
            logging.exception("Exception caught during message delegation.")
            response_body = ErrorSchema.from_exc_info()
        await self._conf.driver.push_rpc_message_out(context, response_body)

    async def delegate_rpc_message(
        self, context: RPCContext[Any], body: Optional[Schema]
    ) -> Union[Schema, Tuple[Schema, ...], None]:
        """Delegate RPC message received server-side."""

        handler_fn, args = self.get_handler_fn(context, body)
        return await handler_fn(*args)

    @overload
    def get_handler_fn(
        self, context: RPCContext[Any], body: None
    ) -> Tuple[InterfaceNoBody[HandlerReturns], Tuple[()]]:
        ...

    @overload
    def get_handler_fn(
        self, context: RPCContext[Any], body: Schema
    ) -> Tuple[Interface[Schema, HandlerReturns], Tuple[Schema]]:
        ...

    def get_handler_fn(
        self, context: RPCContext[Any], body: Optional[Schema]
    ) -> Union[
        Tuple[Interface[Schema, HandlerReturns], Tuple[Schema]],
        Tuple[InterfaceNoBody[HandlerReturns], Tuple[()]],
    ]:
        resource: Resource = getattr(self._conf.impl, context.resource)
        try:
            if body:
                return getattr(resource, context.interface), (body,)
            return getattr(resource, context.interface), ()
        except AttributeError:
            raise exceptions.RPCServerInterfaceNotImplemented(
                resource=context.resource, interface=context.interface
            )



from __future__ import annotations

import copy

from typing import Any
from typing import Dict
from typing import Generic
from typing import Tuple
from typing import Union

from typing_extensions import Protocol
from typing_extensions import runtime_checkable

from .typing import SchemaInContraT
from .typing import SchemaOutCoT
from .typing import TypeVar


__all__ = (
    "App",
    "Boundary",
    "BoundaryT",
    "Resource",
    "Interface",
    "InterfaceNoBody",
)


@runtime_checkable
class Interface(Protocol[SchemaInContraT, SchemaOutCoT]):
    async def __call__(self, body: SchemaInContraT) -> SchemaOutCoT:
        ...


@runtime_checkable
class InterfaceNoBody(Protocol[SchemaOutCoT]):
    async def __call__(self) -> SchemaOutCoT:
        ...


InterfacesT = Dict[str, Union[Interface[Any, Any], InterfaceNoBody[Any]]]


class ResourceMeta(type):
    def __init__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)
        cls.__interfaces__: InterfacesT = {}
        # pylint false positive below: https://github.com/PyCQA/pylint/issues/3268
        cls._collect__interfaces__(namespace)  # pylint: disable=no-value-for-parameter

    def _collect__interfaces__(cls, namespace: Dict[str, Any]) -> None:
        try:
            annotations: Dict[str, Any] = namespace["__annotations__"]
        except KeyError:
            return

        for key, value in annotations.items():
            try:
                is_interface = isinstance(value, Interface)
            except TypeError:
                is_interface = False

            if is_interface:
                cls.__interfaces__[key] = value


class Resource(metaclass=ResourceMeta):
    """App resource specification."""

    def __init__(self) -> None:
        # redeclare _resources since mypy & pylint don't recognize attribute from mcs
        self.__interfaces__: InterfacesT


ResourcesT = Dict[str, Resource]


class BoundaryMeta(type):
    def __init__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)
        cls.__resources__: ResourcesT = {}
        # pylint false positive below: https://github.com/PyCQA/pylint/issues/3268
        cls._collect__resources__(namespace)  # pylint: disable=no-value-for-parameter

    def _collect__resources__(cls, namespace: Dict[str, Any]) -> None:
        for key, value in namespace.items():
            try:
                is_subclass = issubclass(value, Resource)
            except TypeError:
                if isinstance(value, Resource):
                    cls.__resources__[key] = value
            else:
                # convert to instance
                if is_subclass:
                    value = value()
                    cls.__resources__[key] = value


class Boundary(metaclass=BoundaryMeta):
    """App boundary."""

    def __init__(self) -> None:
        # redeclare _resources since mypy & pylint don't recognize attribute from mcs
        self.__resources__: ResourcesT

        # A copy of each resource is made in order to avoid inadvertantly modifying the
        # class attribute. For example, a mixin may choose to attach metadata directly
        # on the resource instances.
        for attr in self.__resources__.keys():
            resource = self.__resources__[attr]
            resource_copy = copy.deepcopy(resource)
            setattr(self, attr, resource_copy)
            self.__resources__[attr] = resource_copy


BoundaryT = TypeVar("BoundaryT", bound=Boundary)


class App(Generic[BoundaryT]):
    def __init__(self, boundary: BoundaryT, **_kwargs: Any) -> None:
        self._boundary = boundary



from __future__ import annotations

import asyncio
import uuid

from dataclasses import dataclass
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from zealous import Schema

from ..base import BaseRPCDriver
from ..base import RPCContext


__all__ = ("AsyncioQRPCDriver",)


class AsyncioQRPCMessageDriverContext:
    def __init__(self) -> None:
        self.uuid = uuid.uuid4().hex


AsyncioQRPCContext = RPCContext[AsyncioQRPCMessageDriverContext]


@dataclass
class MessageIn:
    context: AsyncioQRPCContext
    body: Optional[Schema]


@dataclass
class MessageOut:
    context: AsyncioQRPCContext
    body: Union[Schema, Tuple[Schema, ...], None]


class AsyncioQRPCDriver(BaseRPCDriver[AsyncioQRPCMessageDriverContext]):
    def __init__(self, *, maxsize: int = 0) -> None:
        super().__init__()
        self.maxsize = maxsize

        self.__in_queue: Optional[asyncio.Queue[MessageIn]] = None
        self._out_queues: Dict[str, asyncio.Queue[MessageOut]] = {}

    @property
    def _in_queue(self) -> asyncio.Queue[MessageIn]:
        """Defers creation of asyncio.Queue until used in event loop."""
        if self.__in_queue is None:
            self.__in_queue = asyncio.Queue(maxsize=self.maxsize)
        return self.__in_queue

    def reset(self) -> None:
        self.__in_queue = None
        self._out_queues.clear()

    def make_driver_context(  # pylint: disable=no-self-use
        self,
    ) -> AsyncioQRPCMessageDriverContext:
        return AsyncioQRPCMessageDriverContext()

    async def push_rpc_message_in(
        self, context: AsyncioQRPCContext, body: Optional[Schema]
    ) -> None:
        await self._in_queue.put(MessageIn(context, body))
        self._out_queues[context.driver_context.uuid] = asyncio.Queue(
            maxsize=self.maxsize
        )

    async def pop_rpc_message_in(self,) -> Tuple[AsyncioQRPCContext, Optional[Schema]]:
        message_in = await self._in_queue.get()
        return message_in.context, message_in.body

    async def push_rpc_message_out(
        self,
        context: AsyncioQRPCContext,
        response_body: Union[Schema, Tuple[Schema, ...], None],
    ) -> None:
        await self._out_queues[context.driver_context.uuid].put(
            MessageOut(context, response_body)
        )

    async def pop_rpc_message_out(
        self, context: AsyncioQRPCContext
    ) -> Union[Schema, Tuple[Schema, ...], None]:
        message_out = await self._out_queues[context.driver_context.uuid].get()
        del self._out_queues[context.driver_context.uuid]
        return message_out.body



from __future__ import annotations

from typing import Tuple
from typing import TypeVar
from typing import Union

from .schema import Schema


# General purpose TypeVars
T = TypeVar("T")
CoT = TypeVar("CoT", covariant=True)
ContraT = TypeVar("ContraT", contravariant=True)
ClsT = TypeVar("ClsT")
SelfT = TypeVar("SelfT")

# Schema TypeVars
_SchemaInT_bound = Schema
SchemaInT = TypeVar("SchemaInT", bound=_SchemaInT_bound)
SchemaInCoT = TypeVar("SchemaInCoT", bound=_SchemaInT_bound, covariant=True)
SchemaInContraT = TypeVar("SchemaInContraT", bound=_SchemaInT_bound, contravariant=True)

_SchemaOutT_bound = Union[Schema, Schema, Tuple[Schema, ...], None]
SchemaOutT = TypeVar("SchemaOutT", bound=_SchemaOutT_bound)
SchemaOutCoT = TypeVar("SchemaOutCoT", bound=_SchemaOutT_bound, covariant=True)
SchemaOutContraT = TypeVar(
    "SchemaOutContraT", bound=_SchemaOutT_bound, contravariant=True
)
